"""
The fixtures module provides classes returned by fixtures registered by pytest in snappylapy.

Snappylapy provides the following pytest fixtures.

- expect: Expect
    - Allows for validating various expectations on the test results and do snapshot testing.
- load_snapshot: LoadSnapshot
    - Allows loading from a snapshot created by another test.
"""

from __future__ import annotations

from .expectation_classes import (
    BytesExpect,
    DataframeExpect,
    DictExpect,
    ListExpect,
    ObjectExpect,
    StringExpect,
)
from .models import Settings
from .serialization import (
    BytesSerializer,
    JsonPickleSerializer,
    PandasCsvSerializer,
    Serializer,
    StringSerializer,
)
from snappylapy.constants import DIRECTORY_NAMES
from snappylapy.session import SnapshotSession
from typing import Any, Protocol, TypeVar, overload

T = TypeVar("T")


class _CallableExpectation(Protocol):
    """Protocol for callable expectations to use internally in this module."""

    def __call__(
        self,
        data_to_snapshot: Any,  # noqa: ANN401
        name: str | None = None,
        filetype: str = "snapshot.txt",
    ) -> DictExpect | ListExpect | StringExpect | BytesExpect | DataframeExpect:
        """Call the expectation with the given parameters."""
        ...


class Expect:
    """
    Snapshot testing fixture class.

    Do not instantiate this class directly, instead use the `expect` fixture provided by pytest.
    Use this class as a type hint for the `expect` fixture.

    *Calling expect directly*

    The `expect` fixture can be called directly with a value, and it will automatically select
    the appropriate handler (DictExpect, ListExpect, StringExpect, BytesExpect, DataframeExpect, or ObjectExpect)
    depending on the type of the input. The input type must be resolvable to one of the supported types.

    Example usage
    ---------------
    `test_expect_direct_call.py`
    ```python
    from snappylapy.fixtures import Expect

    def test_expect_direct_call(expect: Expect) -> None:
        # Dict input
        data_dict: dict[str, int] = {"a": 1, "b": 2}
        expect(data_dict).to_match_snapshot()

        # List input
        data_list: list[int] = [1, 2, 3]
        expect(data_list).to_match_snapshot()

        # String input
        data_str: str = "pytest example"
        expect(data_str).to_match_snapshot()

        # Bytes input
        data_bytes: bytes = b"binary"
        expect(data_bytes).to_match_snapshot()

        # DataFrame input (requires pandas)
        import pandas as pd
        df: pd.DataFrame = pd.DataFrame({"x": [1, 2]})
        expect(df).to_match_snapshot()

        # Custom object input (falls back to ObjectExpect)
        class Custom:
            def __init__(self) -> None:
                self.value = 42
        custom_obj = Custom()
        expect(custom_obj).to_match_snapshot()
    ```

    The handler is chosen based on the type of `data_to_snapshot`. If the type is not directly supported,
    it falls back to the generic object handler.
    """

    def __init__(
        self,
        snappylapy_session: SnapshotSession,
        snappylapy_settings: Settings,
    ) -> None:
        """"""  # noqa: D419, blank, since used in doc generation
        self.settings = snappylapy_settings

        self.dict = DictExpect(self.settings, snappylapy_session)
        """DictExpect instance for configuring snapshot testing of dictionaries.

        The instance is callable with the following parameters:

        Parameters
        ----------
        - `data_to_snapshot` (`dict`): The dictionary data to be snapshotted.
        - `name` (`str`, optional): The name of the snapshot".
        - `filetype` (`str`, optional): The file type of the snapshot, by default "dict.json".

        Returns
        -------
        DictExpect
        - `DictExpect`: The instance of the DictExpect class.

        Example
        -------
        `test_fixture_expect_dict.py`
        ```python
        import pytest
        from snappylapy.fixtures import Expect

        def test_expect_dict(expect: Expect) -> None:
            data: dict[str, str] = {"key": "value"}
            expect.dict(data).to_match_snapshot()
            expect.dict(data, name="snapshot_name", filetype="dict.json").to_match_snapshot()
        ```
        """

        self.list = ListExpect(self.settings, snappylapy_session)
        """ListExpect instance for configuring snapshot testing of lists.

        The instance is callable with the following parameters:

        Parameters
        ----------
        - `data_to_snapshot` (`list`): The list data to be snapshotted.
        - `name` (`str`, optional): The name of the snapshot.
        - `filetype` (`str`, optional): The file type of the snapshot, by default "list.json".

        Returns
        -------
        ListExpect
        - `ListExpect`: The instance of the ListExpect class.

        Example
        -------
        `test_fixture_expect_list.py`
        ```python
        import pytest
        from snappylapy.fixtures import Expect

        def test_expect_list(expect: Expect) -> None:
            data: list[int] = [1, 2, 3]
            expect.list(data).to_match_snapshot()
        ```
        """

        self.string = StringExpect(self.settings, snappylapy_session)
        """StringExpect instance for configuring snapshot testing of strings.

        The instance is callable with the following parameters:

        Parameters
        ----------
        - `data_to_snapshot` (`str`): The string data to be snapshotted.
        - `name` (`str`, optional): The name of the snapshot.
        - `filetype` (`str`, optional): The file type of the snapshot, by default "string.txt".

        Returns
        -------
        StringExpect
        - `StringExpect`: The instance of the StringExpect class.

        Example
        -------
        `test_fixture_expect_string.py`
        ```python
        import pytest
        from snappylapy.fixtures import Expect

        def test_expect_string(expect: Expect) -> None:
            data: str = "Hello, World!"
            expect.string(data).to_match_snapshot()
        ```
        """

        self.bytes = BytesExpect(self.settings, snappylapy_session)
        """BytesExpect instance for configuring snapshot testing of bytes.

        The instance is callable with the following parameters:

        Parameters
        ----------
        - `data_to_snapshot` (`bytes`): The bytes data to be snapshotted.
        - `name` (`str`, optional): The name of the snapshot.
        - `filetype` (`str`, optional): The file type of the snapshot, by default "bytes.txt".

        Returns
        -------
        BytesExpect
        - `BytesExpect`: The instance of the BytesExpect class.

        Example
        -------
        `test_fixture_expect_bytes.py`
        ```python
        import pytest
        from snappylapy.fixtures import Expect

        def test_expect_bytes(expect: Expect) -> None:
            data: bytes = b"binary data"
            expect.bytes(data).to_match_snapshot()
        ```
        """

        self.dataframe = DataframeExpect(self.settings, snappylapy_session)
        """DataframeExpect instance for configuring snapshot testing of dataframes.

        The instance is callable with the following parameters:

        Parameters
        ----------
        - `data_to_snapshot` (`pd.DataFrame`): The dataframe data to be snapshotted.
        - `name` (`str`, optional): The name of the snapshot.
        - `filetype` (`str`, optional): The file type of the snapshot, by default "dataframe.json".

        Returns
        -------
        DataframeExpect
        - `DataframeExpect`: The instance of the DataframeExpect class.

        Example
        -------
        `test_fixture_expect_dataframe.py`
        ```python
        import pytest
        import pandas as pd
        from snappylapy.fixtures import Expect

        def test_expect_dataframe(expect: Expect) -> None:
            df: pd.DataFrame = pd.DataFrame({"key": ["value1", "value2"]})
            expect.dataframe(df).to_match_snapshot()
        ```
        """

        self.object = ObjectExpect(self.settings, snappylapy_session)
        """ObjectExpect instance for configuring snapshot testing of generic objects.

        The instance is callable with the following parameters:

        Parameters
        ----------
        - `data_to_snapshot` (`object`): The object data to be snapshotted.
        - `name` (`str`, optional): The name of the snapshot.
        - `filetype` (`str`, optional): The file type of the snapshot, by default "object.json".

        Returns
        -------
        ObjectExpect
        - `ObjectExpect`: The instance of the ObjectExpect class.

        Example
        -------
        `test_fixture_expect_object.py`
        ```python
        import pytest
        from snappylapy.fixtures import Expect

        def test_expect_object(expect: Expect) -> None:
            obj: dict[str, str] = {"key": "value"}
            expect.object(obj).to_match_snapshot()
        ```
        """

    @overload
    def __call__(self, data_to_snapshot: dict, name: str | None = None, filetype: str | None = None) -> DictExpect: ...

    @overload
    def __call__(
        self,
        data_to_snapshot: list[Any],
        name: str | None = None,
        filetype: str | None = None,
    ) -> ListExpect: ...

    @overload
    def __call__(self, data_to_snapshot: str, name: str | None = None, filetype: str | None = None) -> StringExpect: ...

    @overload
    def __call__(
        self,
        data_to_snapshot: bytes,
        name: str | None = None,
        filetype: str | None = None,
    ) -> BytesExpect: ...

    @overload
    def __call__(
        self,
        data_to_snapshot: DataframeExpect.DataFrame,
        name: str | None = None,
        filetype: str | None = None,
    ) -> DataframeExpect: ...

    @overload
    def __call__(
        self,
        data_to_snapshot: Any,  # noqa: ANN401
        name: str | None = None,
        filetype: str | None = None,
    ) -> ObjectExpect: ...

    def __call__(
        self,
        data_to_snapshot: dict | list[Any] | str | bytes | DataframeExpect.DataFrame,
        name: str | None = None,
        filetype: str | None = None,
    ) -> DictExpect | ListExpect | StringExpect | BytesExpect | DataframeExpect | ObjectExpect:
        """Call the fixture with the given parameters. Falls back to object handler for custom objects."""
        kwargs: dict[str, str] = {}
        if name is not None:
            kwargs["name"] = name
        if filetype is not None:
            kwargs["filetype"] = filetype

        type_map: dict[type, _CallableExpectation] = {
            dict: self.dict,
            list: self.list,
            str: self.string,
            bytes: self.bytes,
            DataframeExpect.DataFrame: self.dataframe,
        }

        for typ, func in type_map.items():
            if isinstance(typ, type) and isinstance(data_to_snapshot, typ):
                return func(data_to_snapshot, **kwargs)

        # Check if the object is a pandas DataFrame without importing pandas directly
        if (
            type(data_to_snapshot).__module__.startswith("pandas")
            and type(data_to_snapshot).__name__ == "DataFrame"
            # TODO: Create a protocol class instead that contains all the dependencies we are depending on
        ):
            return self.dataframe(data_to_snapshot, **kwargs)  # type: ignore[arg-type]

        # Fallback: treat custom objects as dicts for snapshotting
        return self.object(data_to_snapshot, **kwargs)

    def _read_snapshot(self) -> bytes:
        """Read the snapshot file."""
        return (self.settings.snapshot_dir / self.settings.filename).read_bytes()

    def _read_test_results(self) -> bytes:
        """Read the test results file."""
        return (self.settings.test_results_dir / self.settings.filename).read_bytes()


class LoadSnapshot:
    """
    Snapshot loading class.

    This class provides methods to load snapshots created by other tests.
    Each method loads and deserializes a specific type of snapshot.
    """

    def __init__(self, settings: Settings) -> None:
        """Do not initialize the LoadSnapshot class directly, should be used through the `load_snapshot` fixture in pytest."""  # noqa: E501
        self.settings = settings
        self._current_dependency_index = 0

    def _read_snapshot(self) -> bytes:
        """Read the snapshot file."""
        if self._current_dependency_index >= len(self.settings.depending_tests):
            msg = (
                f"Attempted to load more dependencies ({self._current_dependency_index + 1}) "
                f"than available ({len(self.settings.depending_tests)}). "
                "Check your test's dependency configuration."
            )
            raise IndexError(msg)
        if not self.settings.depending_tests[self._current_dependency_index].snapshots_base_dir:
            msg = "Depending snapshots base directory is not set."
            raise ValueError(msg)
        return (
            self.settings.depending_tests[self._current_dependency_index].snapshots_base_dir
            / DIRECTORY_NAMES.snapshot_dir_name
            / self.settings.depending_tests[self._current_dependency_index].filename
        ).read_bytes()

    def _load_and_deserialize(self, filename_extension: str, deserializer: Serializer[T]) -> T:
        """Set filename extension, read, deserialize, and increment dependency index."""
        self.settings.depending_tests[self._current_dependency_index].filename_extension = filename_extension
        deserialized_data = deserializer.deserialize(self._read_snapshot())
        self._current_dependency_index += 1
        return deserialized_data

    def dict(self) -> dict[Any, Any]:
        """
        Load dictionary snapshot.

        Use this method to load a dictionary snapshot that was created in a previous test.
        This is useful for reusing test data, isolating dependencies, and verifying integration between components.

        Example usage:
        --------------
        `test_load_snapshot_from_file_dict.py`
        ```python
        import pytest
        from snappylapy.fixtures import LoadSnapshot, Expect

        def create_dict() -> dict[str, int]:
            return {"apples": 3, "bananas": 5}

        def test_save_dict_snapshot(expect: Expect) -> None:
            data: dict[str, int] = create_dict()
            expect(data).to_match_snapshot()

        @pytest.mark.snappylapy(depends=[test_save_dict_snapshot])
        def test_load_snapshot_dict(load_snapshot: LoadSnapshot) -> None:
            data: dict[str, int] = load_snapshot.dict()
            assert data["apples"] == 3
            assert data["bananas"] == 5
        ```
        """
        return self._load_and_deserialize(
            "dict.json",
            JsonPickleSerializer[dict](),
        )

    def list(self) -> list[Any]:
        """
        Load list snapshot.

        Use this method to load a list snapshot that was created in a previous test.
        This is useful for reusing test data, isolating dependencies, and verifying integration between components.

        Example usage:
        --------------
        `test_load_snapshot_from_file_list.py`
        ```python
        import pytest
        from typing import Any
        from snappylapy import LoadSnapshot, Expect

        def transform_data(data: list) -> list:
            return [x * 2 for x in data]

        def next_transformation(data: list) -> list:
            return [x + 1 for x in data]

        def test_transform_data(expect: Expect) -> None:
            data = [1, 2, 3]
            result = transform_data(data)
            expect(result).to_match_snapshot()

        @pytest.mark.snappylapy(depends=[test_transform_data])
        def test_next_transformation(load_snapshot: LoadSnapshot, expect: Expect) -> None:
            data: list[Any] = load_snapshot.list()
            result = next_transformation(data)
            expect(result).to_match_snapshot()
        ```
        """
        return self._load_and_deserialize(
            "list.json",
            JsonPickleSerializer[list[Any]](),
        )

    def string(self) -> str:
        """
        Load string snapshot.

        Use this method to load a string snapshot that was created in a previous test.
        This is useful for reusing test data, isolating dependencies, and verifying integration between components.

        Example usage:
        --------------
        `test_load_snapshot_from_file_string.py`
        ```python
        import pytest
        from snappylapy.fixtures import LoadSnapshot, Expect

        def test_save_string_snapshot(expect: Expect) -> None:
            message: str = "Hello, pytest!"
            expect(message).to_match_snapshot()

        @pytest.mark.snappylapy(depends=[test_save_string_snapshot])
        def test_load_snapshot_string(load_snapshot: LoadSnapshot) -> None:
            data: str = load_snapshot.string()
            assert data == "Hello, pytest!"
        ```
        """
        return self._load_and_deserialize(
            "string.txt",
            StringSerializer(),
        )

    def bytes(self) -> bytes:
        r"""
        Load bytes snapshot.

        Use this method to load a bytes snapshot that was created in a previous test.
        This is useful for reusing test data, isolating dependencies, and verifying integration between components.

        Example usage:
        --------------
        `test_load_snapshot_from_file_bytes.py`
        ```python
        import pytest
        from snappylapy.fixtures import LoadSnapshot, Expect

        def test_save_bytes_snapshot(expect: Expect) -> None:
            data: bytes = b"\x01\x02\x03"
            expect(data).to_match_snapshot()

        @pytest.mark.snappylapy(depends=[test_save_bytes_snapshot])
        def test_load_snapshot_bytes(load_snapshot: LoadSnapshot) -> None:
            data: bytes = load_snapshot.bytes()
            assert data == b"\x01\x02\x03"
        ```
        """
        return self._load_and_deserialize(
            "bytes.txt",
            BytesSerializer(),
        )

    def dataframe(self) -> DataframeExpect.DataFrame:
        """
        Load dataframe snapshot.

        Use this method to load a dataframe snapshot that was created in a previous test.
        This is useful for reusing test data, isolating dependencies, and verifying integration between components.

        Example usage:
        --------------
        `test_load_snapshot_from_file_dataframe.py`
        ```python
        import pytest
        import pandas as pd
        from snappylapy.fixtures import LoadSnapshot, Expect

        def test_save_dataframe_snapshot(expect: Expect) -> None:
            df: pd.DataFrame = pd.DataFrame({"numbers": [1, 2, 3]})
            expect(df).to_match_snapshot()

        @pytest.mark.snappylapy(depends=[test_save_dataframe_snapshot])
        def test_load_snapshot_dataframe(load_snapshot: LoadSnapshot) -> None:
            df: pd.DataFrame = load_snapshot.dataframe()
            assert df["numbers"].sum() == 6
        ```
        """
        return self._load_and_deserialize(
            "dataframe.csv",
            PandasCsvSerializer(),
        )

    def object(self) -> object:
        """
        Load object snapshot.

        Use this method to load a generic object snapshot that was created in a previous test.
        This is useful for reusing test data, isolating dependencies, and verifying integration between components.

        Example usage:
        --------------
        `test_load_snapshot_from_file_object.py`
        ```python
        import pytest
        from snappylapy import Expect, LoadSnapshot

        class Custom:
            def __init__(self) -> None:
                self.value = 42

        def test_save_object_snapshot(expect: Expect) -> None:
            obj = Custom()
            expect(obj).to_match_snapshot()

        @pytest.mark.snappylapy(depends=[test_save_object_snapshot])
        def test_load_snapshot_object(load_snapshot: LoadSnapshot) -> None:
            obj = load_snapshot.object()
            assert hasattr(obj, "value")
            assert obj.value == 42
        ```
        """
        return self._load_and_deserialize(
            "object.json",
            JsonPickleSerializer[object](),
        )



"""Utility functions for handling directories in snappylapy."""
import pathlib
from functools import cached_property
from snappylapy.constants import DIRECTORY_NAMES


def find_directories(directory_names: list[str]) -> list[pathlib.Path]:
    """Find directories with the given names."""
    found_dirs: list[pathlib.Path] = []
    for dir_name in directory_names:
        found_dirs.extend(pathlib.Path().rglob(dir_name))
    return found_dirs


def get_file_paths_from_directories(list_of_directories: list[pathlib.Path]) -> list[pathlib.Path]:
    """Get file paths from directories."""
    list_of_files_to_delete: list[pathlib.Path] = []
    for directory in list_of_directories:
        if not directory.is_dir():
            error_msg = f"{directory} is not a directory."
            raise ValueError(error_msg)
        list_of_files_to_delete.extend(file for file in directory.iterdir() if file.is_file())
    return list_of_files_to_delete


class DirectoryNamesUtil:
    """
    Utility class to handle directory names and operations related to them.

    This class extends the DirectoryNames class to provide methods for finding directories
    and file paths created by snappylapy.
    """

    @cached_property
    def all_directory_names(self) -> list[str]:
        """Get all directory names."""
        return [DIRECTORY_NAMES.snapshot_dir_name, DIRECTORY_NAMES.test_results_dir_name]

    @cached_property
    def all_directories_for_test_results(self) -> list[pathlib.Path]:
        """Get all directories for test results."""
        return find_directories([DIRECTORY_NAMES.test_results_dir_name])

    @cached_property
    def all_directories_for_snapshots(self) -> list[pathlib.Path]:
        """Get all directories for snapshots."""
        return find_directories([DIRECTORY_NAMES.snapshot_dir_name])

    @cached_property
    def all_file_paths_test_results(self) -> list[pathlib.Path]:
        """Get all file paths in the test results directory."""
        directories = find_directories([DIRECTORY_NAMES.test_results_dir_name])
        return get_file_paths_from_directories(directories)

    @cached_property
    def all_file_paths_snapshots(self) -> list[pathlib.Path]:
        """Get all file paths in the snapshot directory."""
        directories = find_directories([DIRECTORY_NAMES.snapshot_dir_name])
        return get_file_paths_from_directories(directories)

    @cached_property
    def all_directories_created_by_snappylapy(self) -> list[pathlib.Path]:
        """Get all directories created by snappylapy."""
        return find_directories(self.all_directory_names)

    @cached_property
    def all_file_paths_created_by_snappylapy(self) -> list[pathlib.Path]:
        """Get all file paths created by snappylapy."""
        return get_file_paths_from_directories(
            self.all_directories_created_by_snappylapy,
        )



"""Snapshot testing and expectations for strings."""

from __future__ import annotations

from .base_snapshot import BaseSnapshot
from snappylapy.serialization import StringSerializer


class StringExpect(BaseSnapshot[str]):
    """Snapshot testing for strings."""

    serializer_class = StringSerializer

    def __call__(self, data_to_snapshot: str, name: str | None = None, filetype: str = "string.txt") -> StringExpect:
        """Prepare a string for snapshot testing."""
        if not hasattr(data_to_snapshot, "encode"):
            msg = f"data_to_snapshot of type {type(data_to_snapshot)} must be a string-like object supporting an 'encode' method."  # noqa: E501
            raise TypeError(msg)
        self._prepare_test(data_to_snapshot, name, filetype)
        return self



"""Create cli using the typer library."""

import re
import typer
import pathlib
import subprocess  # noqa: S404
from enum import Enum
from snappylapy._utils_directories import DirectoryNamesUtil
from snappylapy.constants import DIRECTORY_NAMES

app = typer.Typer(
    no_args_is_help=True,
    help="""
    Welcome to the snappylapy CLI!

    Use these commands to initialize your repository, update or clear test results and snapshots,
    and review differences between your test results and snapshots using the 'diff' command.

    - Run *'init'* to set up your repo for snappylapy.
    - Use *'status'* to see the current status of your snapshots.
    - Use *'update'* to refresh snapshots with the latest test results.
    - Use *'clear'* to remove all test results and snapshots (add --force to skip confirmation).
    - Use *'diff'* to view changes between test results and snapshots in your editor.

    For more details on each command, use --help after the command name.
    """,
)


@app.command()
def init() -> None:
    """
    Run this command to initialize your repository for snappylapy.

    This will add a line to your .gitignore file to ensure test results are not tracked by git.
    """
    # Check if .gitignore exists
    gitignore_path = pathlib.Path(".gitignore")
    if not gitignore_path.exists():
        typer.echo("No .gitignore file found. Creating one.")
        gitignore_path.touch()
    # Check if already in .gitignore
    with gitignore_path.open("r") as file:
        lines = file.readlines()
    regex = re.compile(rf"^{re.escape(DIRECTORY_NAMES.test_results_dir_name)}(/|$)")
    if any(regex.match(line) for line in lines):
        typer.echo("Already in .gitignore.")
        return
    # Add to .gitignore to top of file
    line_to_add = f"# Ignore test results from snappylapy\n{DIRECTORY_NAMES.test_results_dir_name}/\n\n"
    with gitignore_path.open("w") as file:
        file.write(line_to_add)
        file.writelines(lines)
    typer.echo(f"Added {DIRECTORY_NAMES.test_results_dir_name}/ to .gitignore.")


@app.command()
def status() -> None:
    """
    Show the status of all files in the snapshot directory.

    This will list how many files are unchanged, changed, or not found compared to the snapshots.
    """
    status_fetcher = _SnapshotStatusManager()
    status_fetcher.print_status()


@app.command()
def clear(
    force: bool = typer.Option(
        False,
        "--force",
        "-f",
        help="Force deletion without confirmation",
    ),
) -> None:
    """
    Use this command to clear all test results and snapshots created by snappylapy.

    This will recursively delete all files and directories related to test results and snapshots.
    Use --force to skip confirmation.

    This finds and deletes all `__test_results__` and `__snapshots__` directories recursively across the working directory.
    """  # noqa: E501
    list_of_files_to_delete = DirectoryNamesUtil().all_file_paths_created_by_snappylapy
    if not list_of_files_to_delete:
        typer.echo("No files to delete.")
        return
    _print_deletion_summary(
        directories_to_delete=DirectoryNamesUtil().all_directories_created_by_snappylapy,
        list_of_files_to_delete=list_of_files_to_delete,
    )
    if not force and not _user_approved_deletion_in_cli():
        return
    # Delete files
    _delete_files(list_of_files_to_delete)
    typer.echo(f"Deleted {len(list_of_files_to_delete)} files.")


@app.command()
def update() -> None:
    """
    Use this command to update all snapshot files with the latest test results.

    This will overwrite existing snapshots with current test outputs, ensuring your snapshots reflect the latest changes.

    The file contents of any files in any of the `__test_results__` folders will be copied to the corresponding `__snapshots__` folder.
    """  # noqa: E501
    status_fetcher = _SnapshotStatusManager()
    if status_fetcher.get_count_of_test_results_files() == 0:
        typer.echo("No test result files found. Run pytest tests that use snappylapy expectations first.")
        return
    files_to_update = status_fetcher.get_files_to_be_updated()
    count_up_to_date_files = status_fetcher.get_count_of_test_results_files() - len(files_to_update)
    if not files_to_update:
        typer.echo(f"All snapshot files are up to date. {count_up_to_date_files} files are up to date.")
        return

    typer.echo(
        f"Found {len(files_to_update)} files to update."
        + (f" {count_up_to_date_files} files are up to date." if count_up_to_date_files > 0 else ""),
    )
    for test_result_file_path in files_to_update:
        snapshot_file_path = _get_snapshot_path_from_test_result_path(test_result_file_path)
        snapshot_file_path.parent.mkdir(parents=True, exist_ok=True)
        snapshot_file_path.write_bytes(test_result_file_path.read_bytes())
        typer.echo(f"Updated snapshot: {snapshot_file_path}")

    typer.echo("Finished updating snapshots.")


@app.command()
def diff() -> None:
    """
    Show the differences between the test results and the snapshots.

    Opens all of the changed diffs in the Visual Studio Code (VSCode) editor.
    This requires that you have VSCode installed and the `code` command available in your PATH.

    More diff viewers will be supported in the future, please raise a request on github with your needs.
    """
    status_fetcher = _SnapshotStatusManager()
    files_to_diff = status_fetcher.get_files_with_status(FileStatus.CHANGED)
    if not files_to_diff:
        status_fetcher.print_status()
        typer.echo("No files have changed, not opening any diffs.")
        return
    typer.echo(f"Opening diffs for {len(files_to_diff)} changed files.")
    for test_result_file_path in files_to_diff:
        open_diff_viewer(
            test_result_file_path=test_result_file_path,
            snapshot_file_path=_get_snapshot_path_from_test_result_path(test_result_file_path),
        )


def _get_snapshot_path_from_test_result_path(test_result_file_path: pathlib.Path) -> pathlib.Path:
    """Get the snapshot file path from the test result file path."""
    return test_result_file_path.parent.parent / DIRECTORY_NAMES.snapshot_dir_name / test_result_file_path.name


def open_diff_viewer(test_result_file_path: pathlib.Path, snapshot_file_path: pathlib.Path) -> None:
    """Open diff viewer for the given file."""
    success: bool = _try_open_diff(test_result_file_path, snapshot_file_path)
    if not success:
        typer.secho(
            f"Could not open diff tool. Files to compare:\n"
            f"  Test result: {test_result_file_path.resolve()}\n"
            f"  Snapshot:    {snapshot_file_path.resolve()}",
            fg=typer.colors.YELLOW,
        )


class FileStatus(Enum):
    """Enum to represent the status of a file."""

    NOT_FOUND = "not_found"
    CHANGED = "changed"
    UNCHANGED = "unchanged"


def _user_approved_deletion_in_cli() -> bool:
    """Ask user for confirmation before deleting files."""
    typer.secho("\nAre you sure you want to delete all test results and snapshots?", fg=typer.colors.BRIGHT_BLUE)
    response = typer.prompt("Type 'yes' to confirm, anything else to abort.", default="no")
    if response.lower() == "yes":
        typer.echo("User confirmed, deleting files.")
        return True
    typer.echo("Aborted, not deleting any files.")
    return False


def _print_deletion_summary(
    directories_to_delete: list[pathlib.Path],
    list_of_files_to_delete: list[pathlib.Path],
) -> None:
    """Print summary of files and directories to be deleted."""
    typer.echo("Deleting files:")
    for file in list_of_files_to_delete:
        typer.echo(f"- {file}")

    typer.secho(
        f"Deleting {len(list_of_files_to_delete)} files from {len(directories_to_delete)} directories:",
        fg=typer.colors.BRIGHT_BLUE,
    )
    for directory in directories_to_delete:
        typer.echo(f"- {directory}")


def _delete_files(list_of_files_to_delete: list[pathlib.Path]) -> None:
    """Delete files."""
    # Delete files
    for file in list_of_files_to_delete:
        file.unlink()
    # Delete directories
    for dir_name in [
        DIRECTORY_NAMES.test_results_dir_name,
        DIRECTORY_NAMES.snapshot_dir_name,
    ]:
        for root_dir in pathlib.Path().rglob(dir_name):
            root_dir.rmdir()


def _try_open_diff(file1: pathlib.Path, file2: pathlib.Path) -> bool:
    """Try to open diff using available tools, return True if successful."""
    diff_commands: list[list[str]] = [
        ["code", "--diff", str(file1.resolve()), str(file2.resolve())],
        ["code.cmd", "--diff", str(file1.resolve()), str(file2.resolve())],  # Windows alternative
    ]

    for command in diff_commands:
        try:
            subprocess.run(command, check=True, timeout=10)  # noqa: S603 - shell=False and args as list, safe usage
        except subprocess.TimeoutExpired:  # noqa: PERF203
            typer.secho(
                f"Diff tool timed out for command: {' '.join(command)}",
                fg=typer.colors.RED,
            )
            continue
        except (subprocess.CalledProcessError, FileNotFoundError):
            continue
        else:
            return True
    return False


class _SnapshotStatusManager:
    """Various utilities to get various snapshot statuses and information."""

    def _check_file_statuses(
        self,
    ) -> dict[pathlib.Path, FileStatus]:
        """Check the status of files in the snapshot directory."""
        files_test_results = DirectoryNamesUtil().all_file_paths_test_results
        file_statuses: dict[pathlib.Path, FileStatus] = {}
        for file_path in files_test_results:
            snapshot_file = file_path.parent.parent / DIRECTORY_NAMES.snapshot_dir_name / file_path.name
            if not snapshot_file.exists():
                file_statuses[file_path] = FileStatus.NOT_FOUND
            elif snapshot_file.stat().st_size != file_path.stat().st_size:
                # TODO: This is not foolproof, does not catch content swaps and byte flips.
                file_statuses[file_path] = FileStatus.CHANGED
            elif snapshot_file.read_bytes() != file_path.read_bytes():
                # TODO: Expensive call, store hashes instead in a data file.
                file_statuses[file_path] = FileStatus.CHANGED
            else:
                file_statuses[file_path] = FileStatus.UNCHANGED
        return file_statuses

    def get_count_of_test_results_files(self) -> int:
        """Return the count of all test results files."""
        files_test_results = DirectoryNamesUtil().all_file_paths_test_results
        return len(files_test_results)

    def _get_status_counts(self) -> dict[FileStatus, int]:
        """Return a dict with counts of each file status for snapshot files."""
        file_statuses: dict[pathlib.Path, FileStatus] = self._check_file_statuses()
        status_counts: dict[FileStatus, int] = dict.fromkeys(FileStatus, 0)
        for status in file_statuses.values():
            status_counts[status] += 1
        return status_counts

    def get_files_with_status(
        self,
        status: FileStatus,
    ) -> list[pathlib.Path]:
        """Get all files with the desired status."""
        file_statuses: dict[pathlib.Path, FileStatus] = self._check_file_statuses()
        return [file for file, file_status in file_statuses.items() if file_status == status]

    def get_files_to_be_updated(self) -> list[pathlib.Path]:
        """Get all files that need to be updated (not unchanged)."""
        file_statuses: dict[pathlib.Path, FileStatus] = self._check_file_statuses()
        return [file for file, file_status in file_statuses.items() if file_status != FileStatus.UNCHANGED]

    def print_status(self) -> None:
        """Print the current status of the snapshot files to the console."""
        status_counts = self._get_status_counts()
        if not status_counts or sum(status_counts.values()) == 0:
            typer.secho("No test result files found.", fg=typer.colors.YELLOW)
            return

        typer.secho("Snapshot status:", underline=True, bold=True)
        for status in FileStatus:
            typer.echo(f"- {status.value}: {status_counts[status]} file(s)")

        if status_counts[FileStatus.CHANGED] > 0 or status_counts[FileStatus.NOT_FOUND] > 0:
            typer.secho("Some files are changed or missing snapshots.", fg=typer.colors.RED)
        else:
            typer.secho("All files are up to date.", fg=typer.colors.GREEN)


if __name__ == "__main__":
    app()



"""Base class for snapshot testing."""

from __future__ import annotations

import pathlib
import Levenshtein
from abc import ABC, abstractmethod
from snappylapy.models import Settings
from snappylapy.serialization import Serializer
from snappylapy.session import SnapshotSession
from typing import Generic, TypeVar

T = TypeVar("T")

LARGE_DIFF_DISABLE_ASSERTION_REWRITES_ON_DISTANCE = 100
LARGE_DIFF_CHARACTER_THRESHOLD = 10_000


def calculate_difference_two_strings(a: str, b: str) -> int:
    """
    Approximate string difference.

    This is a gross approximation and may not reflect true semantic differences.
    """
    return Levenshtein.distance(a, b, score_cutoff=LARGE_DIFF_DISABLE_ASSERTION_REWRITES_ON_DISTANCE)


class BaseSnapshot(ABC, Generic[T]):
    """Base class for snapshot testing."""

    serializer_class: type[Serializer[T]]

    def __init__(
        self,
        settings: Settings,
        snappylapy_session: SnapshotSession,
    ) -> None:
        """Initialize the base snapshot."""
        self.settings = settings
        self._data: T | None = None
        self.snappylapy_session = snappylapy_session

    @abstractmethod
    def __call__(
        self,
        data_to_snapshot: T,
        name: str | None = None,
        filetype: str = "snapshot.txt",
    ) -> BaseSnapshot[T]:
        """Prepare data for snapshot testing."""

    def to_match_snapshot(self) -> None:
        """Assert test results match the snapshot."""
        if not (self.settings.snapshot_dir / self.settings.filename).exists():
            if not self.settings.snapshot_update:
                error_msg = f"Failed `to_match_snapshot()` expectation. Snapshot file not found: {self.settings.filename}, run 'snappylapy update' command in the terminal, or run pytest with the --snapshot-update flag to create it."  # noqa: E501
                raise FileNotFoundError(error_msg)
            self.snappylapy_session.add_created_snapshot(self.settings.filename)
            self._update_snapshot()
            return

        snapshot_data = self._read_file(self.settings.snapshot_dir / self.settings.filename)
        test_data = self._read_file(self.settings.test_results_dir / self.settings.filename)
        try:
            self.compare_snapshot_data(snapshot_data, test_data)
        except AssertionError as error:
            if self.settings.snapshot_update:
                self.snappylapy_session.add_updated_snapshot(self.settings.filename)
                self._update_snapshot()
            else:
                self.snappylapy_session.add_snapshot_test_failed(self.settings.filename)
                diff_msg = str(error)
                error_msg = (
                    "Failed `to_match_snapshot()` expectation. Snapshot does not match test "
                    "results. Run pytest with the --snapshot-update flag to update the "
                    f"snapshot.\n{diff_msg}"
                )
                raise AssertionError(error_msg)  # noqa: B904
        else:
            self.snappylapy_session.add_snapshot_test_succeeded(self.settings.filename)

    def to_align_with_snapshot(self) -> None:
        """
        Non-deterministic snapshot alignment.

        Data structure specific implementations, can fuzzy match strings or compare
        data structures. The default behavior defined is always accepting the snapshot,
        but fails if the snapshot has not been created yet.
        """
        if not (self.settings.snapshot_dir / self.settings.filename).exists():
            if not self.settings.snapshot_update:
                error_msg = f"Failed `to_align_with_snapshot()` expectation. Snapshot file not found: {self.settings.filename}, run 'snappylapy update' command in the terminal, or run pytest with the --snapshot-update flag to create it."  # noqa: E501
                raise FileNotFoundError(error_msg)
            self.snappylapy_session.add_created_snapshot(self.settings.filename)
            self._update_snapshot()
        try:
            self._to_align_with_snapshot_validation()
        except AssertionError:
            if self.settings.snapshot_update:
                self._update_snapshot()
                self.snappylapy_session.add_updated_snapshot(self.settings.filename)
                return
            self.snappylapy_session.add_snapshot_test_failed(self.settings.filename)
            raise
        else:
            self.snappylapy_session.add_snapshot_test_succeeded(self.settings.filename)

    def _to_align_with_snapshot_validation(self) -> None:
        """
        Validate the alignment with the existing snapshot.

        This is a hook for implementing data-structure-specific validations used by
        :meth:`to_align_with_snapshot`. In the base implementation this method is a
        no-op and never raises, meaning that if the snapshot file exists,
        ``to_align_with_snapshot`` will treat the snapshot as accepted/aligned.

        Subclasses should override this method and raise ``AssertionError`` if the
        alignment/validation fails.
        """
        # TODO: The subclass implementations do not exist yet.

    def compare_snapshot_data(self, snapshot_data: bytes, test_data: bytes) -> None:
        """
        Compare snapshot data with test data.

        Optimizations is needed for large data since difflib used in the pytest_assertion plugin rewrites are very slow
        when there are many diffs in large strings.
        """
        snapshot_data_str = snapshot_data.decode()
        snapshot_data_str = snapshot_data_str.replace("\r\n", "\n").replace("\r", "\n")
        test_data_str = test_data.decode()
        test_data_str = test_data_str.replace("\r\n", "\n").replace("\r", "\n")
        if (
            len(snapshot_data_str) > LARGE_DIFF_CHARACTER_THRESHOLD
            or len(test_data_str) > LARGE_DIFF_CHARACTER_THRESHOLD
        ):
            distance = Levenshtein.distance(
                snapshot_data_str,
                test_data_str,
                score_cutoff=LARGE_DIFF_DISABLE_ASSERTION_REWRITES_ON_DISTANCE,
            )
            if distance < LARGE_DIFF_DISABLE_ASSERTION_REWRITES_ON_DISTANCE:
                assert snapshot_data_str == test_data_str
            if snapshot_data_str != test_data_str:
                msg = (
                    "Snapshots are not matching. Data is too large to show differences. "
                    "Instead use the 'snappylapy diff' command to see the differences."
                )
                raise AssertionError(msg)
        else:
            assert snapshot_data_str == test_data_str

    def _prepare_test(self, data: T, name: str | None, extension: str) -> None:
        """Prepare and save test results."""
        if name is not None:
            self.settings.custom_name = str(name)
        self._data = data
        self.settings.filename_extension = extension
        file_path = self.settings.test_results_dir / self.settings.filename
        file_path.parent.mkdir(parents=True, exist_ok=True)
        self._save_test_results(file_path, data)

    def _update_snapshot(self) -> None:
        """Write test results to the snapshot file."""
        snap_path = self.settings.snapshot_dir / self.settings.filename
        test_path = self.settings.test_results_dir / self.settings.filename
        snap_path.parent.mkdir(parents=True, exist_ok=True)
        snap_path.write_bytes(test_path.read_bytes())

    def _read_file(self, path: pathlib.Path) -> bytes:
        """Read file bytes or return placeholder."""
        return path.read_bytes() if path.exists() else b"<No file>"

    def _save_test_results(self, path: pathlib.Path, data: T) -> None:
        """Save data for test results."""
        data_bin = self.serializer_class().serialize(data)
        path.write_bytes(data_bin)
